import { derive_key } from '@cmdcode/crypto-tools/hd';
import { parse_addr } from '@scrow/tapscript/address';
import { cmd_config } from '../config.js';
import { decode_psbt, encode_psbt } from '@cmdcode/psbt2';
import * as assert from '../assert.js';
import { segwit, taproot } from '@scrow/tapscript/sighash';
import { create_tx, create_vin, decode_tx, encode_tx } from '@scrow/tapscript/tx';
import { parse_desc, parse_descriptor } from './descriptors.js';
import * as CONST from './const.js';
import { Buff } from '@cmdcode/buff';
import { encode_script } from '@scrow/tapscript/script';
const { DUST_LIMIT, MIN_TX_FEE, SAT_MULTI, RANDOM_SORT } = CONST;
export class CoreWallet {
    constructor(client, label, config = {}) {
        this._addrs = new Map();
        this._config = config;
        this._client = client;
        this._label = label;
        this._txfee = null;
    }
    get client() {
        return this._client;
    }
    get info() {
        return this.cmd('getwalletinfo', null, { cache: true });
    }
    get is_created() {
        return this.client.wallets_created
            .then(e => Array.isArray(e) && e.includes(this.label));
    }
    get is_loaded() {
        return this.client.wallets_loaded
            .then(e => Array.isArray(e) && e.includes(this.label));
    }
    get label() {
        return this._label;
    }
    get balance() {
        return this.cmd('getbalance', null, { cache: true })
            .then(bal => Number(bal) * SAT_MULTI);
    }
    get network() {
        return this.client.opt.network;
    }
    get new_address() {
        return this.gen_address();
    }
    get new_scriptkey() {
        return this.new_address.then(e => parse_addr(e).asm);
    }
    get utxos() {
        return this.cmd('listunspent', null, { cache: true })
            .then(e => e.map(x => { return { ...x, sats: Math.round(x.amount * SAT_MULTI) }; }));
    }
    get xprvs() {
        return this.cmd('listdescriptors', true, { cache: true })
            .then(({ descriptors }) => descriptors.map(x => parse_descriptor(x)));
    }
    get xpubs() {
        return this.cmd('listdescriptors', null, { cache: true })
            .then(({ descriptors }) => descriptors.map(x => parse_descriptor(x)));
    }
    get xprv() {
        return new Promise(async (res) => {
            const xprvs = await this.xprvs;
            const wpkh = xprvs.find(e => e.keytype === 'wpkh');
            if (wpkh === undefined) {
                throw new Error('unable to locate wpkh descriptor');
            }
            res(wpkh.keystr);
        });
    }
    get xpub() {
        return new Promise(async (res) => {
            const xpubs = await this.xpubs;
            const wpkh = xpubs.find(e => e.keytype === 'wpkh');
            if (wpkh === undefined) {
                throw new Error('unable to locate wpkh descriptor');
            }
            res(wpkh.keystr);
        });
    }
    _debug(...msg) {
        if (this.client.opt.debug) {
            console.log('[wallet]', ...msg);
        }
    }
    _log(...msg) {
        if (this.client.opt.verbose) {
            console.log('[wallet]', ...msg);
        }
    }
    async _create() {
        const payload = { wallet_name: this.label, ...this._config };
        const res = await this.client.cmd('createwallet', payload);
        const err = (res.warning !== undefined && res.warning !== '');
        if (err || res.name !== this.label) {
            throw new Error(`Wallet failed to create: ${JSON.stringify(res, null, 2)}`);
        }
    }
    async _load() {
        const res = await this.client.cmd('loadwallet', this.label);
        if (res.warning !== undefined || res.name !== this.label) {
            throw new Error(`Wallet failed to load: ${JSON.stringify(res, null, 2)}`);
        }
    }
    async init() {
        const info = await this.info;
        if (info.paytxfee === 0) {
            const newtxfee = MIN_TX_FEE / SAT_MULTI;
            await this.cmd('settxfee', newtxfee);
        }
    }
    async load() {
        if (!await this.is_loaded) {
            if (!await this.is_created) {
                await this._create();
            }
            else {
                await this._load();
            }
        }
    }
    async cmd(method, args = [], config) {
        const conf = cmd_config(config);
        conf.params = [...conf.params, `-rpcwallet=${this.label}`];
        return this.client.cmd(method, args, conf);
    }
    async gen_address(config) {
        return this.cmd('getnewaddress', config);
    }
    async get_address(label) {
        let addr = this._addrs.get(label);
        if (addr !== undefined) {
            this._debug('[wallet] using saved address:', addr);
            return addr;
        }
        try {
            const addr_book = await this.cmd('getaddressesbylabel', label, { cache: true });
            const addr_list = Object.keys(addr_book);
            addr = addr_list[0];
        }
        catch {
            addr = await this.gen_address({ label });
        }
        this._addrs.set(label, addr);
        return addr;
    }
    async parse_address(address) {
        return this.cmd('getaddressinfo', address, { cache: true });
    }
    async send_funds(amount, address, mine_block = false) {
        const amt = amount / SAT_MULTI;
        const config = { address, amount: amt, estimate_mode: 'economical' };
        const txid = await this.cmd('sendtoaddress', config);
        if (mine_block)
            await this.client.mine_blocks(1);
        return txid;
    }
    async ensure_funds(min_bal) {
        const bal = await this.balance;
        if (bal <= min_bal && this.label !== 'faucet') {
            await this.drain_faucet(min_bal);
            if (this.network === 'regtest') {
                await this.client.mine_blocks(1);
            }
        }
    }
    async drain_faucet(amount, address) {
        if (address === undefined) {
            address = await this.new_address;
        }
        const faucet = this.client.core.faucet;
        const balance = await faucet.balance;
        if (balance <= amount + 10000) {
            throw new Error('faucet is broke!');
        }
        else {
            return faucet.send_funds(amount, address);
        }
    }
    async get_xprv(label) {
        const xprvs = await this.xprvs;
        return xprvs.find(e => label === e.label);
    }
    async get_signer(desc) {
        const { parent_label, fullpath, keytype } = parse_desc(desc);
        const xprv = await this.get_xprv(parent_label);
        assert.ok(xprv?.extkey?.seckey !== undefined);
        const { seckey: master_key, code } = xprv.extkey;
        const { seckey, pubkey } = derive_key(fullpath, master_key, code, true);
        assert.ok(seckey !== null);
        const signer = (keytype === 'tr') ? taproot : segwit;
        const sign_tx = (txdata, config) => {
            return signer.sign_tx(seckey, txdata, config);
        };
        return { pubkey, sign_tx };
    }
    async create_txout(amount, address) {
        if (typeof address !== 'string') {
            address = await this.new_address;
        }
        return {
            value: BigInt(amount),
            scriptPubKey: parse_addr(address).asm
        };
    }
    async select_utxos(amount, sorter = RANDOM_SORT) {
        const selected = [];
        let total = 0;
        const utxos = await this.utxos;
        utxos.sort(sorter);
        for (const utxo of utxos) {
            selected.push(utxo);
            total += utxo.sats;
            if (total === amount ||
                total > amount + DUST_LIMIT) {
                return selected;
            }
        }
        throw new Error('Insufficient funds!');
    }
    async get_fund_data(templ, config = {}, txfee = 1000) {
        const vin = [], vout = [];
        const txdata = create_tx(templ);
        const vamt = txdata.vout.reduce((prev, curr) => Number(curr.value) + prev, 0);
        const utxos = await this.select_utxos(vamt + txfee);
        const total = utxos.reduce((prev, curr) => curr.sats + prev, 0);
        const change_sats = BigInt(total - vamt - txfee);
        vout.push({
            value: change_sats,
            scriptPubKey: await this.new_scriptkey
        });
        for (let i = 0; i < utxos.length; i++) {
            const { desc, txid, vout, sats, scriptPubKey } = utxos[i];
            const { pubkey, sign_tx } = await this.get_signer(desc);
            const prevout = { value: sats, scriptPubKey };
            const txinput = create_vin({ txid, vout, prevout });
            const txconfig = { sigflag: 0x81, pubkey, txinput };
            const signature = sign_tx(txdata, { ...txconfig, ...config });
            const witness = [signature.hex];
            if (desc.startsWith('wpkh'))
                witness.push(pubkey.hex);
            vin.push({ ...txinput, witness });
        }
        return { vin, vout };
    }
    async fund_tx(templ, config = {}, txfee = 1000) {
        const txdata = create_tx(templ);
        const funds = await this.get_fund_data(templ, config, txfee);
        txdata.vin.push(...funds.vin);
        txdata.vout.push(...funds.vout);
        return txdata;
    }
    async fund_psbt(psbt, options, txfee) {
        const pdata = decode_psbt(Buff.base64(psbt).hex);
        const tx_idx = pdata.global.findIndex(e => e.label === 'UNSIGNED_TX');
        if (tx_idx === -1) {
            throw new Error('no tx data found in psbt');
        }
        const txdata = decode_tx(pdata.global[tx_idx].value, false);
        const funds = await this.get_fund_data(txdata, options, txfee);
        for (const txin of funds.vin) {
            assert.exists(txin.prevout);
            const prevout = Buff.join([
                Buff.big(txin.prevout.value, 8).reverse(),
                encode_script(txin.prevout.scriptPubKey)
            ]);
            const witness = Buff.join([
                '02',
                encode_script(txin.witness[0]),
                encode_script(txin.witness[1])
            ]);
            pdata.inputs.push([
                { label: 'WITNESS_UTXO', value: prevout.hex },
                { label: 'FINAL_SCRIPTWITNESS', value: witness.hex }
            ]);
        }
        txdata.vin.push(...funds.vin);
        txdata.vout.push(...funds.vout);
        pdata.global[tx_idx].value = encode_tx(txdata, false).hex;
        return encode_psbt(pdata);
    }
}
//# sourceMappingURL=wallet.js.map